from xshinnosuke.layers import Conv2D, Flatten, Dense, Embedding
from xshinnosuke.models import Module
from xshinnosuke.nn.optimizers import Adam
from xshinnosuke.nn import Variable, CrossEntropy
from xshinnosuke.utils import DataSet, DataLoader
import cupy
import os


class CNN(Module):
    def __init__(self, n_classes=4):
        super().__init__()
        self.mean = None
        self.std = None
        self.height = None
        self.width = None
        self.epochs = 50
        self.embed = Embedding(input_dim=15, output_dim=1)
        self.conv1 = Conv2D(32, kernel_size=3, use_bias=True, padding=1, activation='relu')
        self.conv2 = Conv2D(64, kernel_size=3, use_bias=True, padding=1, activation='relu')
        self.conv3 = Conv2D(128, kernel_size=3, use_bias=True, padding=1, activation='relu')
        self.flat = Flatten()
        self.fc1 = Dense(1024, activation='relu')
        self.fc2 = Dense(512, activation='relu')
        self.fc3 = Dense(n_classes)

    def training(self, x, y, height, width):
        self.height = height
        self.width = width
        self.train()
        x, y = self.preprocess_data(x, y)
        train_dataset = DataSet(x, y)
        train_loader = DataLoader(train_dataset, batch_size=100, shuffle=True)
        optimizer = Adam(self.parameters())
        criterion = CrossEntropy()
        for epoch in range(self.epochs):
            for xs, ys in train_loader:
                xs = Variable(xs, dtype='int')
                ys = Variable(ys)
                optimizer.zero_grad()
                pred = self(xs)
                loss = criterion(pred, ys)
                loss.backward()
                optimizer.step()
                print(f'epoch: {epoch} loss: {loss.data}')
        save_path = os.path.join(os.path.dirname(__file__), 'model')
        self.save(save_path)

    def prediction(self, x, height, width):
        self.height = height
        self.width = width
        model_path = os.path.join(os.path.dirname(__file__), 'model')
        self.load(model_path)
        x = cupy.array(x)
        x = self.convert_x(x)
        x = Variable(x, dtype='int')
        out = self.predict(x)
        return cupy.argmax(out.data, 1).tolist()

    def preprocess_data(self, x, y):
        x = cupy.array(x)
        y = cupy.array(y)
        x = self.convert_x(x)
        return x[None, :], y[:, None]

    def convert_x(self, x):
        for i in range(1, 15):
            target = 2 ** i
            pos = cupy.where(x == target)
            x[pos] = i
        return x

    def normalize(self, x):
        self.mean = cupy.mean(x)
        self.std = cupy.std(x)
        return (x - self.mean) / self.std

    def forward(self, x, *args):
        x = self.embed(x)
        x = x.view(-1, 1, self.height, self.width)
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.flat(x)
        x = self.fc1(x)
        x = self.fc2(x)
        x = self.fc3(x)
        return x



from .utils import load_mnist
from .network import FCNet, ConvNet
from xshinnosuke.utils import DataSet, DataLoader
from xshinnosuke.nn import Variable, CrossEntropy
from xshinnosuke.nn.optimizers import SGD


def go():
    #########################  EPOCH BATCH SIZE
    EPOCH = 5
    BATCH_SIZE = 256

    #########################  Read Data
    trainset, valset, testset = load_mnist()

    #########################  Network
    net = FCNet()

    #########################  Compile
    net.compile(optimizer='sgd', loss='cross_entropy')

    #########################  Train FC
    net.fit(trainset[0], trainset[1], batch_size=BATCH_SIZE, epochs=EPOCH, validation_data=valset)

    #########################  Evaluate FC
    train_acc, train_loss = net.evaluate(trainset[0], trainset[1], batch_size=500)
    val_acc, val_loss = net.evaluate(valset[0], valset[1], batch_size=500)
    test_acc, test_loss = net.evaluate(testset[0], testset[1], batch_size=500)
    print('#' * 10, 'Fully Connected Network')
    print(f'Train acc -> {train_acc}, Train loss -> {train_loss}')
    print(f'Val acc -> {val_acc}, Val loss -> {val_loss}')
    print(f'Test acc -> {test_acc}, Test loss -> {test_loss}')

    print('Start Training ConvNet---')
    #########################  Network
    net = ConvNet()

    #########################  Compile
    net.compile(optimizer='sgd', loss='cross_entropy')

    #########################  Train Conv
    net.fit(trainset[0], trainset[1], batch_size=BATCH_SIZE, epochs=EPOCH, validation_data=valset)

    train_acc, train_loss = net.evaluate(trainset[0], trainset[1], batch_size=500)
    val_acc, val_loss = net.evaluate(valset[0], valset[1], batch_size=500)
    test_acc, test_loss = net.evaluate(testset[0], testset[1], batch_size=500)
    print('#' * 10, 'Convolutional Network')
    print(f'Train acc -> {train_acc}, Train loss -> {train_loss}')
    print(f'Val acc -> {val_acc}, Val loss -> {val_loss}')
    print(f'Test acc -> {test_acc}, Test loss -> {test_loss}')



from .global_graph import np
from typing import Tuple


def im2col(inputs: np.ndarray, out_h: int, out_w: int, kernel_h: int, kernel_w: int, stride: Tuple):
    batch_nums, n_C_prev, n_H_prev, n_W_prev = inputs.shape
    col = np.zeros((batch_nums, n_C_prev, kernel_h, kernel_w, out_h, out_w))

    for y in range(kernel_h):
        y_max = y + stride[0] * out_h
        for x in range(kernel_w):
            x_max = x + stride[1] * out_w
            col[:, :, y, x, :, :] = inputs[:, :, y:y_max:stride[0], x:x_max:stride[1]]

    col = col.transpose((0, 4, 5, 1, 2, 3)).reshape(batch_nums * out_h * out_w, -1)
    return col


def col2im(inputs_shape: tuple, pad_size: int, kernel_h: int, kernel_w: int, stride: Tuple, dcol: np.ndarray):
    batch_nums, n_C_prev, n_H_prev, n_W_prev = inputs_shape  # 填充前的shape
    n_H = (n_H_prev + 2 * pad_size - kernel_h) // stride[0] + 1
    n_W = (n_W_prev + 2 * pad_size - kernel_w) // stride[1] + 1

    dcol = dcol.reshape((batch_nums, n_H, n_W, n_C_prev, kernel_h, kernel_w)).transpose(0, 3, 4, 5, 1, 2)

    output = np.zeros(
        (batch_nums, n_C_prev, n_H_prev + 2 * pad_size + stride[0] - 1, n_W_prev + 2 * pad_size + stride[1] - 1))

    for y in range(kernel_h):
        y_max = y + stride[0] * n_H
        for x in range(kernel_w):
            x_max = x + stride[1] * n_W
            output[:, :, y:y_max:stride[0], x:x_max:stride[1]] += dcol[:, :, y, x, :, :]

    return output[:, :, pad_size:n_H_prev + pad_size, pad_size:n_W_prev + pad_size]


def initialize_ops_grad(*ops):
    # 初始化梯度
    for op in ops:
        if op is not None and op.requires_grad:
            op.zero_grad()



from flask import Flask
from flask import jsonify
from flask import render_template
from flask import request
import numpy as np
from .model import CNN
import os


app = Flask(__name__)
net = CNN()
pre_arrs = None
file_dir_path = os.path.dirname(__file__)


@app.route("/")
def hello():
    return render_template('index.html')


@app.route("/record", methods=['POST'])
def move():
    message = request.form
    key = message['key']
    h = message['height']
    w = message['width']
    arrs = message['array'][1:-1]
    with open(f'{file_dir_path}/train_{h}_{w}.txt', 'a') as f1:
        f1.write(arrs)
        f1.write('\n')

    with open(f'{file_dir_path}/label_{h}_{w}.txt', 'a') as f1:
        f1.write(key)
        f1.write('\n')

    return jsonify({'status': 'OK'})


@app.route("/train", methods=['POST'])
def train():
    message = request.form
    h = int(message['height'])
    w = int(message['width'])
    x = []
    y = []
    with open(f'{file_dir_path}/train_{h}_{w}.txt', 'r') as f1:
        for line in f1.readlines():
            line = list(map(int, line.strip().split(',')))
            x.append(line)

    with open(f'{file_dir_path}/label_{h}_{w}.txt', 'r') as f2:
        for target in f2.readlines():
            y.append(int(target))

    net.training(x, y, h, w)
    return jsonify({'status': 'OK'})


@app.route("/ai", methods=['POST'])
def ai():
    global pre_arrs
    message = request.form
    arrs = list(map(int, message['array'][1:-1].split(',')))
    h = int(message['height'])
    w = int(message['width'])

    if pre_arrs is not None and np.equal(arrs, pre_arrs).all():
        key = int(np.random.choice([0, 1, 2, 3]))
    else:
        key = net.prediction(arrs, h, w)[0]

    pre_arrs = arrs

    return jsonify({'status': 'OK', 'key': key})


def go():
    app.run(debug=True)





#coding=utf-8
'''
这是主程序，主要定义了Gobang和Robot两个对象：
1.Gobang主要是图像界面交互的实现。
2.Robot是基于五子棋的一些基本规则写出来的一个简单智能程序，不包含
神经网络的搭建。
'''

from tkinter import *
from tkinter.ttk import *
from tkinter.filedialog import askopenfilename
from tkinter.filedialog import asksaveasfile
from .network import CNN
from .utils import *


class GoBang:
    def __init__(self):
        self.someoneWin = False
        self.humanChessed = False
        self.IsStart = False
        self.player = 0
        self.playmethod = 0
        self.bla_start_pos = [235, 235]
        self.whi_chessed = []
        self.bla_chessed = []
        self.board = self.init_board()
        self.window = Tk()
        self.var = IntVar()
        self.var.set(0)
        self.var1 = IntVar()
        self.var1.set(0)
        self.window.title("myGoBang")
        self.window.geometry("600x470+80+80")
        self.window.resizable(0, 0)
        self.can = Canvas(self.window, bg="#EEE8AC", width=470, height=470)
        self.draw_board()
        self.can.grid(row=0, column=0)
        self.net_board = self.get_net_board()
        self.robot = Robot(self.board)
        self.sgf = SGFflie()
        self.net = CNN()

    @staticmethod
    def init_board():
        list1 = [[-1]*15 for i in range(15)]
        return list1

    def draw_board(self):
        for row in range(15):
            if row == 0 or row == 14:
                self.can.create_line((25, 25 + row * 30), (445, 25 + row * 30), width=2)
            else:
                self.can.create_line((25, 25 + row * 30), (445, 25 + row * 30), width=1)
        for col in range(15):
            if col == 0 or col == 14:
                self.can.create_line((25 + col * 30, 25), (25 + col * 30, 445), width=2)
            else:
                self.can.create_line((25 + col * 30, 25), (25 + col * 30, 445), width=1)
        self.can.create_oval(112, 112, 118, 118, fill="black")
        self.can.create_oval(352, 112, 358, 118, fill="black")
        self.can.create_oval(112, 352, 118, 358, fill="black")
        self.can.create_oval(232, 232, 238, 238, fill="black")
        self.can.create_oval(352, 352, 358, 358, fill="black")

    def get_nearest_po(self, x, y):
        flag = 600
        position = ()
        for point in self.net_board:
            distance = get_distance([x, y], point)
            if distance < flag:
                flag = distance
                position = point
        return position

    def no_in_chessed(self, pos):
        whi_chess = self.check_chessed(pos, self.whi_chessed)
        bla_chess = self.check_chessed(pos, self.bla_chessed)
        return whi_chess == False and bla_chess == False

    def ai_no_in_chessed(self, pos, value):
        no_in_chessed = self.no_in_chessed(pos)
        return no_in_chessed and value < 4000

    @staticmethod
    def check_chessed(point, chessed):
        if len(chessed) == 0:
            return False
        flag = 0
        for p in chessed:
            if point[0] == p[0] and point[1] == p[1]:
                flag = 1
        if flag == 1:
            return True
        else:
            return False

    def have_five(self, chessed):
        if len(chessed) == 0:
            return False
        for row in range(15):
            for col in range(15):
                x = 25 + row * 30
                y = 25 + col * 30
                if self.check_chessed((x, y), chessed) and \
                                self.check_chessed((x, y + 30), chessed) and \
                                self.check_chessed((x, y + 60), chessed) and \
                                self.check_chessed((x, y + 90), chessed) and \
                                self.check_chessed((x, y + 120), chessed):
                    return True
                elif self.check_chessed((x, y), chessed) and \
                                self.check_chessed((x + 30, y), chessed) and \
                                self.check_chessed((x + 60, y), chessed) and \
                                self.check_chessed((x + 90, y), chessed) and \
                                self.check_chessed((x + 120, y), chessed):
                    return True
                elif self.check_chessed((x, y), chessed) and \
                                self.check_chessed((x + 30, y + 30), chessed) and \
                                self.check_chessed((x + 60, y + 60), chessed) and \
                                self.check_chessed((x + 90, y + 90), chessed) and \
                                self.check_chessed((x + 120, y + 120), chessed):
                    return True
                elif self.check_chessed((x, y), chessed) and \
                                self.check_chessed((x + 30, y - 30), chessed) and \
                                self.check_chessed((x + 60, y - 60), chessed) and \
                                self.check_chessed((x + 90, y - 90), chessed) and \
                                self.check_chessed((x + 120, y - 120), chessed):
                    return True
                else:
                    pass
        return False

    def check_win(self):
        if self.have_five(self.whi_chessed):
            label = Label(self.window, text="White Win!", background='#FFF8DC', font=("yahei", 15, "bold"))
            label.place(relx=0, rely=0, x=480, y=40)
            return True
        elif self.have_five(self.bla_chessed):
            label = Label(self.window, text="Black Win!", background='#FFF8DC', font=("yahei", 15, "bold"))
            label.place(relx=0, rely=0, x=480, y=40)
            return True
        else:
            return False

    def draw_chessed(self):
        if len(self.whi_chessed) != 0:
            for tmp in self.whi_chessed:
                oval = pos_to_draw(*tmp[0:2])
                self.can.create_oval(oval, fill="white")

        if len(self.bla_chessed) != 0:
            for tmp in self.bla_chessed:
                oval = pos_to_draw(*tmp[0:2])
                self.can.create_oval(oval, fill="black")

    def draw_a_chess(self, x, y, player=None):
        _x, _y = pos_in_qiju(x, y)
        oval = pos_to_draw(x, y)

        if player == 0:
            self.can.create_oval(oval, fill="black")
            self.bla_chessed.append([x, y, 0])
            self.board[_x][_y] = 1
        elif player == 1:
            self.can.create_oval(oval, fill="white")
            self.whi_chessed.append([x, y, 1])
            self.board[_x][_y] = 0
        else:
            print(AttributeError("please select player"))
        return

    def AIrobotChess(self):
        cnn_predict = self.net.prediction(self.board)
        if self.player % 2 == 0:
            if len(self.bla_chessed) == 0 and len(self.whi_chessed) == 0:
                self.draw_a_chess(*self.bla_start_pos, 0)
            else:
                _x, _y, _ = self.robot.MaxValue_po(1, 0)
                newPoint = pos_in_board(_x, _y)
                if self.ai_no_in_chessed(cnn_predict, _):
                    self.draw_a_chess(*cnn_predict, 0)
                else:
                    self.draw_a_chess(*newPoint, 0)

        else:
            self.robotChess()

    def robotChess(self):
        if self.player == 0:
            if len(self.bla_chessed) == 0 and len(self.whi_chessed) == 0:
                self.draw_a_chess(*self.bla_start_pos, player=0)
                return
            else:
                _x, _y, _ = self.robot.MaxValue_po(0, 1)
                newPoint = pos_in_board(_x, _y)
                self.draw_a_chess(*newPoint, player=0)
        else:#白棋下
            _x, _y, _ = self.robot.MaxValue_po(1, 0)
            newPoint = pos_in_board(_x, _y)
            self.draw_a_chess(*newPoint, player=1)

    def chess(self, event):
        if self.someoneWin is True or self.IsStart is False:
            return

        ex = event.x
        ey = event.y
        if not click_in_board(ex, ey):
            return

        neibor_po = self.get_nearest_po(ex, ey)
        if self.no_in_chessed(neibor_po):
            if self.player == 3:
                total_chessed = len(self.bla_chessed)\
                                + len(self.whi_chessed)
                if (total_chessed % 2) == 0:
                    self.draw_a_chess(*neibor_po, 0)
                else:
                    self.draw_a_chess(*neibor_po, 1)
                self.someoneWin = self.check_win()
            else:
                if self.player == 0:
                    self.draw_a_chess(*neibor_po, 1)
                else:
                    self.draw_a_chess(*neibor_po, 0)

                self.someoneWin = self.check_win()
                if self.playmethod == 0:
                    self.AIrobotChess()
                else:
                    self.robotChess()
                self.someoneWin = self.check_win()

    @staticmethod
    def get_net_board():
        net_list = []
        for row in range(15):
            for col in range(15):
                point = pos_in_board(row, col)
                net_list.append(point)
        return net_list

    def resetButton(self):
        self.someoneWin = False
        self.IsStart = False
        self.whi_chessed.clear()
        self.bla_chessed.clear()
        self.board = self.init_board()
        self.robot = Robot(self.board)
        label = Label(self.window,
                      text="          ",
                      background="#F0F0F0",
                      font=("yahei", 15, "bold"))
        label.place(relx=0, rely=0, x=480, y=40)
        self.can.delete("all")
        self.draw_board()
        self.can.grid(row=0, column=0)

    def BakcAChess(self):
        if self.someoneWin == False:
            if len(self.whi_chessed) != 0:
                p = self.whi_chessed.pop()
                x, y = pos_in_qiju(*p[0:2])
                self.board[x][y] = -1

            if self.player == 0 and len(self.bla_chessed) != 1:
                p = self.bla_chessed.pop()
                x, y = pos_in_qiju(*p[0:2])
                self.board[x][y] = -1

            elif self.player == 1 and len(self.bla_chessed) != 0:
                p = self.bla_chessed.pop()
                x, y = pos_in_qiju(*p[0:2])
                self.board[x][y] = -1

            else:
                pass

            self.can.delete("all")
            self.draw_board()
            self.draw_chessed()

    def startButton(self):
        if not self.IsStart:
            self.IsStart = True
            if self.player % 2 == 0:
                if self.playmethod == 0:
                    self.AIrobotChess()
                elif self.playmethod == 1:
                    self.robotChess()
                self.draw_chessed()

    def selectColor(self):
        if not self.IsStart:
            if self.var.get() == 0:
                self.player = 0
            elif self.var.get() == 1:
                self.player = 1
            elif self.var.get() == 3:
                self.player = 3
            else:
                pass

        return

    def selectMathod(self):
        if self.IsStart == False:
            if self.var1.get() == 0:
                self.playmethod = 0
            elif self.var1.get() == 1:
                self.playmethod = 1
            else:
                pass
        return

    def createqipu(self):
        qipu = []
        step = 0
        totalstep = len(self.whi_chessed) + len(self.bla_chessed)
        while step < totalstep:
            if totalstep == 0:
                break
            flag = int(step / 2)
            if step % 2 == 0:
                pos = pos_in_qiju(*self.bla_chessed[flag][0:2])
                qipu.append([*pos, 0, step + 1])
            else:
                pos = pos_in_qiju(*self.whi_chessed[flag][0:2])
                qipu.append([*pos, 1, step + 1])
            step += 1
        return qipu

    def OpenFile(self):
        file_path = askopenfilename(filetypes=(('sgf file', '*.sgf'),
                                                    ('All File', '*.*')))
        if len(file_path) == 0:
            return

        qipu = self.sgf.openfile(file_path)

        self.whi_chessed.clear()
        self.bla_chessed.clear()

        for point in qipu:
            pos = pos_in_board(*point[0:2])

            if point[2] == 0:
                self.bla_chessed.append([*pos, 0])
            else:
                self.whi_chessed.append([*pos, 1])

        self.can.delete("all")
        self.draw_board()
        self.draw_chessed()

    def SaveFile(self, method=1):
        qipu = self.createqipu()
        if method == 0:
            try:
                file = asksaveasfile(filetypes=(('sgf file', '*.sgf'),
                                                ('All File', '*.*')))
                file.close()
            except AttributeError:
                return

            pathName = file.name
            newName = pathName + '.sgf'
            os.rename(pathName, newName)
            f = open(newName, 'w')
            data = self.sgf.createdata(qipu)
            f.write(data)
            f.close()

        elif method == 1:
            self.sgf.savefile(qipu)

    def start(self):
        b3 = Button(self.window, text="start", command=self.startButton)
        b3.place(relx=0, rely=0, x=495, y=100)

        b1 = Button(self.window, text="reset", command=self.resetButton)
        b1.place(relx=0, rely=0, x=495, y=150)

        b2 = Button(self.window, text="withdraw", command=self.BakcAChess)
        b2.place(relx=0, rely=0, x=495, y=200)

        b4 = Radiobutton(self.window, text="choose white", variable=self.var, value=0, command=self.selectColor)
        b4.place(relx=0, rely=0, x=495, y=250)

        b5 = Radiobutton(self.window, text="choose black", variable=self.var, value=1, command=self.selectColor)
        b5.place(relx=0, rely=0, x=495, y=280)

        b6 = Button(self.window, text="open chessbook", command=self.OpenFile)
        b6.place(relx=0, rely=0, x=495, y=400)

        b7 = Button(self.window, text="save chessbook", command=self.SaveFile)
        b7.place(relx=0, rely=0, x=495, y=430)

        b8 = Radiobutton(self.window, text="ai", variable=self.var1, value=0, command=self.selectMathod)
        b8.place(relx=0, rely=0, x=490, y=340)

        b9 = Radiobutton(self.window, text="normal", variable=self.var1, value=1, command=self.selectMathod)
        b9.place(relx=0, rely=0, x=490, y=370)

        b10 = Radiobutton(self.window, text="p2p", variable=self.var, value=3, command=self.selectColor)
        b10.place(relx=0, rely=0, x=495, y=310)

        self.can.bind("<Button-1>", lambda x: self.chess(x))
        self.window.mainloop()


def go():
    game = GoBang()
    game.start()
    del game



from .utils import SGFflie
import os
import xshinnosuke
from xshinnosuke.models import Sequential
from xshinnosuke.nn import Variable
from xshinnosuke.layers import Conv2D, ReLU, MaxPooling2D, Flatten, Dense, Dropout, Reshape


class CNN:
    def __init__(self, keep_prob=0.5):
        self.sgf = SGFflie()
        # static graph
        self.net = Sequential(
            Reshape((1, 15, 15), input_shape=(225, )),
            Conv2D(out_channels=32, kernel_size=5, use_bias=True, padding=2, kernel_initializer='he_normal',
                   bias_initializer=xshinnosuke.nn.Matrix(0.1)),
            ReLU(inplace=True),
            MaxPooling2D(kernel_size=2),
            Conv2D(out_channels=64, kernel_size=5, use_bias=True, padding=2),
            ReLU(inplace=True),
            MaxPooling2D(kernel_size=2),
            Flatten(),
            Dense(out_features=1024),
            ReLU(inplace=True),
            Dropout(keep_prob),
            Dense(225)
        )
        self.net.compile(optimizer='adam', loss='sparse_cross_entropy', lr=1e-4)

    def forward(self, x):
        return self.net.forward(x)

    def training(self, data_path, save_path, epochs=1, batch_size=None):
        path = self.sgf.allFileFromDir(data_path)
        for filepath in path:
            x, y = self.sgf.createTraindataFromqipu(filepath)
            self.net.fit(x, y, batch_size=batch_size, epochs=epochs)
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        self.restore_save(save_path, 0)

    def prediction(self, qiju):
        _qiju = self.createdataformqiju(qiju)
        pred = self.net.predict(Variable(_qiju[0]))
        point = [0, 0]
        l = pred.argmax(axis=1).data
        for i in range(15):
            if ((i + 1) * 15) > l:
                point[0] = int(i*30 + 25)
                point[1] = int((l - i * 15) * 30 + 25)
                break
        return point

    @staticmethod
    def createdataformqiju(qiju):
        data = []
        tmp = []
        for row in qiju:
            for point in row:
                if point == -1:
                    tmp.append(0.0)
                elif point == 0:
                    tmp.append(2.0)
                elif point == 1:
                    tmp.append(1.0)
        data.append(tmp)
        return data

    def restore_save(self, path, method=1):
        # read
        if method == 1:
            self.net.load(os.path.join(path, 'net'))
            print('finish reading')
        elif method == 0:
            self.net.save(os.path.join(path, 'net'))
            print('finish saving')



from xshinnosuke.models import Sequential
from xshinnosuke.layers import Dense, Flatten, MaxPooling2D, Conv2D, BatchNormalization, ReLU, Reshape


def FCNet(n_classes=10):
    return Sequential(
        Dense(512, activation='relu', input_shape=(784, )),
        Dense(256, activation='relu'),
        Dense(n_classes)
    )


def ConvNet(n_classes=10):
    return Sequential(
        Reshape((1, 28, 28), input_shape=(784, )),
        Conv2D(8, 3),
        BatchNormalization(),
        ReLU(True),
        MaxPooling2D(),
        Conv2D(16, 3),
        BatchNormalization(),
        ReLU(True),
        MaxPooling2D(),
        Conv2D(32, 3),
        BatchNormalization(),
        ReLU(True),
        Flatten(),
        Dense(100, activation='relu'),
        Dense(n_classes)
    )



from ..nn.core import Layer, Variable
from ..nn.global_graph import np
from functools import reduce
from ..nn.initializers import get_initializer
from ..nn.functional import dropout2d, batchnorm, layernorm, groupnorm
from ..nn.grad_fn import Dropout2DBackward, BatchnormBackward, LayernormBackward, GroupnormBackward


class Dropout(Layer):
    def __init__(self, keep_prob: float):
        # prob :probability of keeping a unit active.
        self.keep_prob = keep_prob
        self.mask = None
        super(Dropout, self).__init__()

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            output = dropout2d(inbound, self.keep_prob)
            # output是一个Variable
            return output
        super(Dropout, self).__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        self.data = dropout2d(self.input_data, self.keep_prob, is_training)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        Dropout2DBackward(self.data)


class BatchNormalization(Layer):
    def __init__(self, epsilon=1e-6, momentum=0.99, axis=1, gamma_initializer='ones', beta_initializer='zeros',
                 moving_mean_initializer='zeros', moving_variance_initializer='ones', **kwargs):
        # axis=1 when input Fully Connected Layers(data shape:(M,N),where M donotes Batch-size,and N represents feature nums)  ---also axis=-1 is the same
        # axis=1 when input Convolution Layers(data shape:(M,C,H,W),represents Batch-size,Channels,Height,Width,respectively)
        self.epsilon = epsilon
        self.axis = axis
        self.momentum = momentum
        self.gamma_initializer = get_initializer(gamma_initializer)
        self.beta_initializer = get_initializer(beta_initializer)
        self.moving_mean_initializer = get_initializer(moving_mean_initializer)
        self.moving_variance_initializer = get_initializer(moving_variance_initializer)
        self.moving_mean = None
        self.moving_variance = None
        self.cache = None
        super(BatchNormalization, self).__init__(**kwargs)

    def initial_params(self, input_shape=None):
        if input_shape is not None:
            self.input_shape = input_shape
        assert len(self.input_shape) >= 1
        n_in = self.input_shape[self.axis - 1]
        gamma = self.gamma_initializer(n_in, name='xs_variable')
        beta = self.beta_initializer(n_in, name='xs_variable')
        self.variables.append(gamma)
        self.variables.append(beta)
        self.moving_mean = self.moving_mean_initializer(n_in)
        self.moving_variance = self.moving_variance_initializer(n_in)

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            if len(self.variables) == 0:
                self.initial_params(inbound.shape[1:])
            output = batchnorm(inbound, self.variables[0], self.variables[1], self.axis, self.epsilon, self.momentum,
                               self.moving_mean, self.moving_variance)
            return output
        super().__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x

        self.data = batchnorm(self.input_data, self.variables[0], self.variables[1], self.axis, self.epsilon,
                              self.momentum, self.moving_mean, self.moving_variance, is_training)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        BatchnormBackward(self.data)


class LayerNormalization(Layer):
    def __init__(self, epsilon=1e-10, gamma_initializer='ones', beta_initializer='zeros'):
        self.epsilon = epsilon
        self.gamma_initializer = get_initializer(gamma_initializer)
        self.beta_initializer = get_initializer(beta_initializer)
        self.shape_field = None
        self.cache = None
        super(LayerNormalization, self).__init__()

    def initial_params(self, input_shape=None):
        gamma = self.gamma_initializer(self.input_shape, name='xs_variable')
        beta = self.beta_initializer(self.input_shape, name='xs_variable')
        self.variables.append(gamma)
        self.variables.append(beta)

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            if len(self.variables) == 0:
                self.initial_params(inbound.shape[1:])
            output = layernorm(inbound, self.variables[0], self.variables[1], self.epsilon)
            return output
        Layer.__call__(self, inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x

        self.data = layernorm(self.input_data, self.variables[0], self.variables[1], self.epsilon, is_training)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        LayernormBackward(self.data)


class GroupNormalization(Layer):
    def __init__(self, epsilon=1e-5, groups=16, gamma_initializer='ones', beta_initializer='zeros'):
        self.epsilon = epsilon
        self.G = groups
        self.gamma_initializer = get_initializer(gamma_initializer)
        self.beta_initializer = get_initializer(beta_initializer)
        self.shape_field = None
        self.cache = None
        super(GroupNormalization, self).__init__()

    def initial_params(self, input_shape=None):
        c = self.input_shape[0]
        assert c % self.G == 0
        gamma = self.gamma_initializer((1, c, 1, 1), name='xs_variable')
        beta = self.beta_initializer((1, c, 1, 1), name='xs_variable')
        self.variables.append(gamma)
        self.variables.append(beta)

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            if len(self.variables) == 0:
                self.initial_params(inbound.shape[1:])
            output = groupnorm(inbound, self.variables[0], self.variables[1], self.epsilon, self.G)
            return output
        Layer.__call__(self, inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x

        self.data = groupnorm(self.input_data, self.variables[0], self.variables[1], self.epsilon, self.G, is_training)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        GroupnormBackward(self)



from .core import Variable, Constant
from .functional import *
from .initializers import *
from .objectives import MeanSquaredError, MeanAbsoluteError, SparseCrossEntropy, CrossEntropy, BinaryCrossEntropy, get_objective

name = 'xshinnosuke-nn'



import os
import time
import math


class Robot:
    def __init__(self, _board):
        self.board = _board

    def haveValuePoints(self, player, enemy, board):
        points = []

        for x in range(15):
            for y in range(15):
                list1 = []
                list2 = []
                list3 = []
                list4 = []
                if self.board[x][y] == -1:
                    for tmp in range(9):
                        i = x + tmp - 4
                        j = y + tmp - 4
                        if i < 0 or i > 14:
                            list1.append(-2)
                        else:
                            list1.append(board[i][y])
                        if j < 0 or j > 14:
                            list2.append(-2)
                        else:
                            list2.append(board[x][j])
                        if i < 0 or j < 0 or i > 14 or j > 14:
                            list3.append(-2)
                        else:
                            list3.append(board[i][j])
                        k = y - tmp + 4
                        if i < 0 or k < 0 or i > 14 or k > 14:
                            list4.append(-2)
                        else:
                            list4.append(board[i][k])


                    playerValue = self.value_point(player, enemy, list1, list2, list3, list4)
                    enemyValue = self.value_point(enemy, player, list1, list2, list3, list4)
                    if enemyValue >= 10000:
                        enemyValue -= 500
                    elif enemyValue >= 5000:
                        enemyValue -= 300
                    elif enemyValue >= 2000:
                        enemyValue -= 250
                    elif enemyValue >= 1500:
                        enemyValue -= 200
                    elif enemyValue >= 99:
                        enemyValue -= 10
                    elif enemyValue >= 5:
                        enemyValue -= 1
                    value = playerValue + enemyValue
                    if value > 0:
                        points.append([x, y, value])
        return points

    def MaxValue_po(self, player, enemy):
        points = self.haveValuePoints(player, enemy, self.board)
        flag = 0
        _point = []
        for p in points:
            if p[2] > flag:
                _point = p
                flag = p[2]
        return _point[0], _point[1], _point[2]

    def value_point(self, player, enemy, list1, list2, list3, list4):

        flag = 0
        flag += self.willbefive(player, list1)
        flag += self.willbefive(player, list2)
        flag += self.willbefive(player, list3)
        flag += self.willbefive(player, list4)
        flag += self.willbealive4(player, list1)
        flag += self.willbealive4(player, list2)
        flag += self.willbealive4(player, list3)
        flag += self.willbealive4(player, list4)
        flag += self.willbesleep4(player, enemy, list1)
        flag += self.willbesleep4(player, enemy, list2)
        flag += self.willbesleep4(player, enemy, list3)
        flag += self.willbesleep4(player, enemy, list4)
        flag += self.willbealive3(player, list1)
        flag += self.willbealive3(player, list2)
        flag += self.willbealive3(player, list3)
        flag += self.willbealive3(player, list4)
        flag += self.willbesleep3(player, enemy, list1)
        flag += self.willbesleep3(player, enemy, list2)
        flag += self.willbesleep3(player, enemy, list3)
        flag += self.willbesleep3(player, enemy, list4)
        flag += self.willbealive2(player, enemy, list1)
        flag += self.willbealive2(player, enemy, list2)
        flag += self.willbealive2(player, enemy, list3)
        flag += self.willbealive2(player, enemy, list4)
        flag += self.willbesleep2(player, enemy, list1)
        flag += self.willbesleep2(player, enemy, list2)
        flag += self.willbesleep2(player, enemy, list3)
        flag += self.willbesleep2(player, enemy, list4)
        return flag

    @staticmethod
    def willbefive(player, checklist):
        if checklist[0] == player and checklist[1] == player and \
                checklist[2] == player and checklist[3] == player:
            return 10000
        elif checklist[5] == player and checklist[6] == player and \
                checklist[7] == player and checklist[8] == player:
            return 10000
        elif checklist[2] == player and checklist[3] == player and \
                checklist[5] == player and checklist[6] == player:
            return 10000
        elif checklist[1] == player and checklist[2] == player and \
                checklist[3] == player and checklist[5] == player:
            return 10000
        elif checklist[3] == player and checklist[5] == player and \
                checklist[6] == player and checklist[7] == player:
            return 10000
        else:
            return 0

    @staticmethod
    def willbealive4(player, checklist):
        if checklist[0] == -1 and checklist[1] == player and \
                checklist[2] == player and checklist[3] == player \
                and checklist[5] == -1:
            return 5000
        elif checklist[3] == -1 and checklist[5] == player and \
                checklist[6] == player and checklist[7] == player \
                and checklist[8] == -1:
            return 5000
        elif checklist[1] == -1 and checklist[2] == player and \
                checklist[3] == player and checklist[5] == player \
                and checklist[6] == -1:
            return 5000
        elif checklist[2] == -1 and checklist[3] == player and \
                checklist[5] == player and checklist[6] == player \
                and checklist[7] == -1:
            return 5000
        else:
            return 0

    @staticmethod
    def willbesleep4(player, enemy, checklist):
        if checklist[0] == enemy and checklist[1] == player and \
                checklist[2] == player and checklist[3] == player \
                and checklist[5] == -1:
            return 1700
        elif checklist[1] == enemy and checklist[2] == player and \
                checklist[3] == player and checklist[5] == player \
                and checklist[6] == -1:
            return 1700
        elif checklist[2] == enemy and checklist[3] == player and \
                checklist[5] == player and checklist[6] == player \
                and checklist[7] == -1:
            return 1700
        elif checklist[3] == enemy and checklist[5] == player and \
                checklist[6] == player and checklist[7] == player \
                and checklist[8] == -1:
            return 1700
        elif checklist[0] == -1 and checklist[1] == player and \
                checklist[2] == player and checklist[3] == player \
                and checklist[5] == enemy:
            return 1700
        elif checklist[1] == -1 and checklist[2] == player and \
                checklist[3] == player and checklist[5] == player \
                and checklist[6] == enemy:
            return 1700
        elif checklist[2] == -1 and checklist[3] == player and \
                checklist[5] == player and checklist[6] == player \
                and checklist[7] == enemy:
            return 1700
        elif checklist[3] == -1 and checklist[5] == player and \
                checklist[6] == player and checklist[7] == player \
                and checklist[8] == enemy:
            return 1700
        else:
            return 0

    @staticmethod
    def willbealive3(player, checklist):
        if checklist[0] == -1 and checklist[1] == -1 and \
                checklist[2] == player and checklist[3] == player \
                and checklist[5] == -1:
            return 1900
        elif checklist[1] == -1 and checklist[2] == -1 and \
                checklist[3] == player and checklist[5] == player \
                and checklist[6] == -1:
            return 1900
        elif checklist[2] == -1 and checklist[3] == -1 and \
                checklist[5] == player and checklist[6] == player \
                and checklist[7] == -1:
            return 1900
        elif checklist[1] == -1 and checklist[2] == player and \
                checklist[3] == player and checklist[5] == -1 \
                and checklist[6] == -1:
            return 1900
        elif checklist[2] == -1 and checklist[3] == player and \
                checklist[5] == player and checklist[6] == -1 \
                and checklist[7] == -1:
            return 1900
        elif checklist[3] == -1 and checklist[5] == player and \
                checklist[6] == player and checklist[7] == -1 \
                and checklist[8] == -1:
            return 1900
        elif checklist[0] == -1 and checklist[1] == player and \
                checklist[2] == player and checklist[3] == -1 \
                and checklist[5] == -1:
            return 1600
        elif checklist[2] == -1 and checklist[3] == player and \
                checklist[6] == player and checklist[5] == -1 \
                and checklist[7] == -1:
            return 1600
        elif checklist[3] == -1 and checklist[5] == player and \
                checklist[7] == player and checklist[6] == -1 \
                and checklist[8] == -1:
            return 1600
        elif checklist[3] == -1 and checklist[5] == -1 and \
                checklist[7] == player and checklist[6] == player \
                and checklist[8] == -1:
            return 1600
        elif checklist[0] == -1 and checklist[1] == player and \
                checklist[2] == player and checklist[3] == -1 \
                and checklist[6] == -1:
            return 1600
        elif checklist[0] == -1 and checklist[1] == player and \
                checklist[2] == player and checklist[3] == -1 \
                and checklist[6] == -1:
            return 1600
        else:
            return 0

    @staticmethod
    def willbesleep3(player, enemy, checklist):
        if checklist[1] == enemy and checklist[2] == player and \
                checklist[3] == player and checklist[5] == -1 \
                and checklist[6] == -1:
            return 350
        elif checklist[2] == enemy and checklist[3] == player and \
                checklist[5] == player and checklist[6] == -1 \
                and checklist[7] == -1:
            return 350
        elif checklist[3] == enemy and checklist[5] == player and \
                checklist[6] == player and checklist[7] == -1 \
                and checklist[8] == -1:
            return 350
        elif checklist[0] == -1 and checklist[1] == -1 and \
                checklist[2] == player and checklist[3] == player \
                and checklist[5] == enemy:
            return 350
        elif checklist[1] == -1 and checklist[2] == -1 and \
                checklist[3] == player and checklist[5] == player \
                and checklist[6] == enemy:
            return 350
        elif checklist[2] == -1 and checklist[3] == -1 and \
                checklist[5] == player and checklist[6] == player \
                and checklist[7] == enemy:
            return 350
        elif checklist[0] == enemy and checklist[1] == -1 and \
                checklist[2] == player and checklist[3] == player \
                and checklist[5] == -1 and checklist[6] == enemy:
            return 300
        elif checklist[1] == enemy and checklist[2] == -1 and \
                checklist[3] == player and checklist[5] == player \
                and checklist[6] == -1 and checklist[7] == enemy:
            return 300
        elif checklist[2] == enemy and checklist[3] == -1 and \
                checklist[5] == player and checklist[6] == player \
                and checklist[7] == -1 and checklist[8] == enemy:
            return 300
        elif checklist[0] == enemy and checklist[1] == player and \
                checklist[2] == -1 and checklist[3] == player \
                and checklist[5] == -1 and checklist[6] == enemy:
            return 300
        elif checklist[1] == enemy and checklist[2] == player and \
                checklist[3] == -1 and checklist[5] == player \
                and checklist[6] == -1 and checklist[7] == enemy:
            return 300
        elif checklist[2] == enemy and checklist[3] == player and \
                checklist[5] == -1 and checklist[6] == player \
                and checklist[7] == -1 and checklist[8] == enemy:
            return 300
        elif checklist[0] == enemy and checklist[1] == player and \
                checklist[2] == -1 and checklist[3] == player \
                and checklist[5] == -1 and checklist[6] == enemy:
            return 300
        elif checklist[1] == enemy and checklist[2] == player and \
                checklist[3] == -1 and checklist[5] == player \
                and checklist[6] == -1 and checklist[7] == enemy:
            return 300
        elif checklist[3] == enemy and checklist[5] == -1 and \
                checklist[6] == player and checklist[7] == player \
                and checklist[8] == -1:
            return 300
        elif checklist[0] == enemy and checklist[1] == player and \
                checklist[2] == player and checklist[3] == -1 \
                and checklist[5] == -1:
            return 300
        elif checklist[2] == enemy and checklist[3] == player and \
                checklist[5] == -1 and checklist[6] == player \
                and checklist[7] == -1:
            return 300
        elif checklist[3] == enemy and checklist[5] == player and \
                checklist[6] == -1 and checklist[7] == player \
                and checklist[8] == -1:
            return 300
        elif checklist[0] == player and checklist[1] == player and \
                checklist[2] == -1 and checklist[3] == -1 \
                and checklist[5] == enemy:
            return 300
        elif checklist[2] == enemy and checklist[3] == player and \
                checklist[5] == -1 and checklist[6] == -1 \
                and checklist[7] == player:
            return 300
        elif checklist[3] == enemy and checklist[5] == player and \
                checklist[6] == -1 and checklist[7] == -1 \
                and checklist[8] == player:
            return 300
        elif checklist[0] == player and checklist[1] == -1 and \
                checklist[2] == -1 and checklist[3] == player \
                and checklist[5] == enemy:
            return 300
        elif checklist[1] == player and checklist[2] == -1 and \
                checklist[3] == -1 and checklist[5] == player \
                and checklist[6] == enemy:
            return 300
        elif checklist[3] == enemy and checklist[5] == -1 and \
                checklist[6] == -1 and checklist[7] == player \
                and checklist[8] == player:
            return 300
        elif checklist[0] == -1 and checklist[1] == player and \
                checklist[2] == player and checklist[3] == -1 \
                and checklist[5] == enemy:
            return 30
        elif checklist[2] == -1 and checklist[3] == player and \
                checklist[5] == -1 and checklist[6] == player \
                and checklist[7] == enemy:
            return 300
        elif checklist[3] == -1 and checklist[5] == player and \
                checklist[6] == -1 and checklist[7] == player \
                and checklist[8] == enemy:
            return 300
        elif checklist[0] == -1 and checklist[1] == player and \
                checklist[2] == -1 and checklist[3] == player \
                and checklist[5] == enemy:
            return 300
        elif checklist[1] == -1 and checklist[2] == player and \
                checklist[3] == -1 and checklist[5] == player \
                and checklist[6] == enemy:
            return 300
        elif checklist[3] == -1 and checklist[5] == -1 and \
                checklist[6] == player and checklist[7] == player \
                and checklist[8] == enemy:
            return 300
        elif checklist[0] == player and checklist[1] == -1 and \
                checklist[2] == player and checklist[3] == -1 \
                and checklist[5] == enemy:
            return 300
        elif checklist[1] == enemy and checklist[2] == player and \
                checklist[3] == -1 and checklist[5] == -1 \
                and checklist[6] == player:
            return 300
        elif checklist[2] == player and checklist[3] == -1 and \
                checklist[5]== -1 and checklist[6] == player \
                and checklist[7] == enemy:
            return 300
        elif checklist[3] == enemy and checklist[5] == -1 and \
                checklist[6] == player and checklist[7] == -1 \
                and checklist[8] == player:
            return 300
        else:
            return 0

    @staticmethod
    def willbealive2(player, enemy, checklist):
        if checklist[1] == -1 and checklist[2] == -1 and \
                checklist[3] == player and checklist[5] == -1 \
                and checklist[6] == -1:
            return 99
        elif checklist[2] == -1 and checklist[3] == -1 and \
                checklist[5] == player and checklist[6] == -1 \
                and checklist[7] == -1:
            return 99
        elif checklist[0] == -1 and checklist[1] == -1 and \
                checklist[2] == -1 and checklist[3] == player \
                and checklist[5] == -1 and checklist[6] == enemy:
            return 99
        elif checklist[1] == -1 and checklist[2] == -1 and \
                checklist[3] == -1 and checklist[5] == player \
                and checklist[6] == -1 and checklist[7] == enemy:
            return 99
        elif checklist[1] == enemy and checklist[2] == -1 and \
                checklist[3] == player and checklist[5] == -1 \
                and checklist[6] == -1 and checklist[7] == -1:
            return 99
        elif checklist[2] == enemy and checklist[3] == -1 and \
                checklist[5] == player and checklist[6] == -1 \
                and checklist[7] == -1 and checklist[8] == -1:
            return 99
        else:
            return 0

    @staticmethod
    def willbesleep2(player, enemy, checklist):
        if checklist[2] == enemy and checklist[3] == player and \
                checklist[5] == -1 and checklist[6] == -1 \
                and checklist[7] == -1:
            return 5
        elif checklist[3] == enemy and checklist[5] == player and \
                checklist[6] == -1 and checklist[7] == -1 \
                and checklist[8] == -1:
            return 5
        elif checklist[0] == -1 and checklist[1] == -1 and \
                checklist[2] == -1 and checklist[3] == player \
                and checklist[5] == enemy:
            return 5
        elif checklist[1] == -1 and checklist[2] == -1 and \
                checklist[3] == -1 and checklist[5] == player \
                and checklist[6] == enemy:
            return 5
        elif checklist[1] == enemy and checklist[2] == -1 and \
                checklist[3] == player and checklist[5] == -1 \
                and checklist[6] == -1 and checklist[7] == enemy:
            return 5
        elif checklist[2] == enemy and checklist[3] == -1 and \
                checklist[5] == player and checklist[6] == -1 \
                and checklist[7] == -1 and checklist[8] == enemy:
            return 5
        elif checklist[0] == enemy and checklist[1] == -1 and \
                checklist[2] == player and checklist[3] == -1 \
                and checklist[5] == -1 and checklist[6] == enemy:
            return 5
        elif checklist[2] == enemy and checklist[3] == -1 and \
                checklist[5] == -1 and checklist[6] == player \
                and checklist[7] == -1 and checklist[8] == enemy:
            return 5
        elif checklist[0] == enemy and checklist[1] == -1 and \
                checklist[2] == -1 and checklist[3] == player \
                and checklist[5] == -1 and checklist[6] == enemy:
            return 5
        elif checklist[1] == enemy and checklist[2] == -1 and \
                checklist[3] == -1 and checklist[5] == player \
                and checklist[6] == -1 and checklist[7] == enemy:
            return 5
        elif checklist[0] == -1 and checklist[1] == player and \
                checklist[2] == -1 and checklist[3] == -1 \
                and checklist[5] == enemy:
            return 5
        elif checklist[3] == -1 and checklist[5] == -1 and \
                checklist[6] == -1 and checklist[7] == player \
                and checklist[8] == enemy:
            return 5
        elif checklist[0] == -1 and checklist[1] == -1 and \
                checklist[2] == player and checklist[3] == -1 \
                and checklist[5] == enemy:
            return 5
        elif checklist[2] == -1 and checklist[3] == -1 and \
                checklist[5] == -1 and checklist[6] == player \
                and checklist[7] == enemy:
            return 5
        elif checklist[1] == enemy and checklist[2] == player and \
                checklist[3] == -1 and checklist[5] == -1 \
                and checklist[6] == -1:
            return 5
        elif checklist[3] == enemy and checklist[5] == -1 and \
                checklist[6] == player and checklist[7] == -1 \
                and checklist[8] == -1:
            return 5
        elif checklist[0] == enemy and checklist[1] == player and \
                checklist[2] == -1 and checklist[3] == -1 \
                and checklist[5] == -1:
            return 5
        elif checklist[3] == enemy and checklist[5] == -1 and \
                checklist[6] == -1 and checklist[7] == player \
                and checklist[8] == -1:
            return 5
        else:
            return 0


class SGFflie:
    def __init__(self, save_path='./save/', train_path='./train'):
        self.POS = 'abcdefghijklmno'
        self.savepath = save_path
        self.trainpath = train_path

    def openfile(self, filepath):
        f = open(filepath, 'r')
        data = f.read()
        f.close()

        effective_data = data.split(';')
        s = effective_data[2:-1]

        board = []
        step = 0
        for point in s:
            x = self.POS.index(point[2])
            y = self.POS.index(point[3])
            color = step % 2
            step += 1
            board.append([x, y, color, step])

        return board

    def savefile(self, board):
        data = self.createdata(board)
        if not os.path.exists(self.savepath):
            os.makedirs(self.savepath)
        filepath = os.path.join(self.savepath, data.split(';')[1] + ".sgf")
        f = open(filepath, 'w')
        f.write(data)
        f.close()
        return

    def createdata(self, board):
        now = time.localtime(time.time())
        _time = ''
        for index in range(6):
            _time = _time + str(now[index])
        data = '(;' + _time + ";"

        for it in board:
            if it[2] == 0:
                data = data + 'B[' + self.POS[it[0]] + self.POS[it[1]] + "];"
            else:
                data = data + 'W[' + self.POS[it[0]] + self.POS[it[1]] + "];"
        data = data + ')'
        return data

    def createTraindataFromqipu(self, path, color=0):
        qipu = self.openfile(path)

        bla = qipu[::2]
        whi = qipu[1::2]
        bla_step = len(bla)
        whi_step = len(whi)

        train_x = []
        train_y = []

        if color == 0:
            temp_x = [0.0 for i in range(225)]
            for index in range(bla_step):
                _x = [0.0 for i in range(225)]
                _y = [0.0 for i in range(225)]
                if index == 0:
                    train_x.append(_x)
                    _y[bla[index][0]*15 + bla[index][1]] = 2.0
                    train_y.append(_y)
                else:
                    _x = temp_x.copy()
                    train_x.append(_x)
                    _y[bla[index][0] * 15 + bla[index][1]] = 2.0
                    train_y.append(_y)

                temp_x[bla[index][0] * 15 + bla[index][1]] = 2.0
                if index < whi_step:
                    temp_x[whi[index][0] * 15 + whi[index][1]] = 1.0
        return train_x, train_y

    def createTraindataFromqipu1(self, path, color=0):
        qipu = self.openfile(path)

        bla = qipu[::2]
        whi = qipu[1::2]
        bla_step = len(bla)
        whi_step = len(whi)

        train_x = []
        train_y = []

        if color == 0:
            temp_x = [[[0.0, 0.0, 0.0] for j in range(15)] for k in range(15)]
            for index in range(bla_step):
                _x = [[[0.0, 0.0, 0.0] for j in range(15)] for k in range(15)]
                _y = [0.0 for i in range(225)]
                if index == 0:
                    train_x.append(_x)
                    _y[bla[index][0]*15 + bla[index][1]] = 1.0
                    train_y.append(_y)
                else:
                    _x = temp_x.copy()
                    train_x.append(_x)
                    _y[bla[index][0] * 15 + bla[index][1]] = 1.0
                    train_y.append(_y)

                temp_x[bla[index][0]][bla[index][1]][1] = 1.0
                if index < whi_step:
                    temp_x[whi[index][0]][whi[index][1]][2] = 1.0
        for tmp in train_x:
            for x in tmp:
                for y in x:
                    if y[1] == 0 and y[2] == 0:
                        y[0] = 1
        return train_x, train_y

    def createTraindata(self):
        filepath = self.allFileFromDir(self.savepath)
        train_x = []
        train_y = []
        for path in filepath:
            x, y = self.createTraindataFromqipu(path)
            train_x = train_x + x
            train_y = train_y + y
        return train_x, train_y

    @staticmethod
    def allFileFromDir(Dirpath):
        pathDir = os.listdir(Dirpath)
        pathfile = []
        for allDir in pathDir:
            child = os.path.join('%s%s' % (Dirpath, allDir))
            pathfile.append(child)
        return pathfile

    def createqijuFromqipu(self, path):
        qipu = self.openfile(path)
        bla = qipu[::2]
        whi = qipu[1::2]
        qiju = [[-1]*15 for i in range(15)]

        for tmp in bla:
            qiju[tmp[0]][tmp[1]] = -2
        for tmp in whi:
            qiju[tmp[0]][tmp[1]] = -7
        return qiju


def get_distance(p0, p1):
    return math.sqrt((p0[0] - p1[0]) ** 2 + (p0[1] - p1[1]) ** 2)


def pos_in_board(x, y):
    return x * 30 + 25, y * 30 + 25


def pos_in_qiju(x, y):
    return int((x - 25) / 30), int((y - 25) / 30)


def pos_to_draw(*args):
    x, y = args
    return x - 11, y - 11, x + 11, y + 11


def click_in_board(x, y):
    return 10 < x < 460 and 10 < y < 460



from ..nn.core import Layer, Variable
from xshinnosuke.nn.initializers import get_initializer
from ..nn.functional import embedding
from ..nn.grad_fn import EmbeddingBackward
from ..nn import global_graph as GlobalGraph


class Embedding(Layer):
    def __init__(self, input_dim, output_dim, embeddings_initializer='uniform', mask_zero=False, **kwargs):
        super(Embedding, self).__init__(**kwargs)
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.initializer = get_initializer(embeddings_initializer)
        self.mask_zero = mask_zero

    def initial_params(self, input_shape=None):
        w = self.initializer((self.input_dim, self.output_dim), name='xs_variable')
        self.variables.append(w)

    def compute_output_shape(self, input_shape=None):
        return input_shape + (self.output_dim,)

    def __call__(self, inbound):
        assert self.output_dim is not None and self.input_dim is not None
        if isinstance(inbound, Variable):
            if GlobalGraph.inputs is None:
                GlobalGraph.inputs = inbound

            if len(self.variables) == 0:
                self.initial_params()
            output = embedding(inbound, self.variables[0])
            return output
        super(Embedding, self).__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        assert self.input_data.data.ndim == 2

        self.data = embedding(self.input_data, self.variables[0])
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        EmbeddingBackward(self)



import gzip
from six.moves import urllib
import os
import platform
import pickle
import time


start_time = None
file_dir_path = os.path.dirname(__file__)


def format_size(bytes):
    try:
        bytes = float(bytes)
        kb = bytes / 1024
    except:
        print("Wrong format")
        return "Error"
    if kb >= 1024:
        M = kb / 1024
        if M >= 1024:
            G = M / 1024
            return "%.3fG" % G
        else:
            return "%.3fM" % M
    else:
        return "%.3fK" % kb


def schedule(blocknum, blocksize, totalsize):
    global start_time
    speed = (blocknum * blocksize) / (time.time() - start_time)
    speed_str = " Speed: %s" % format_size(speed)
    recv_size = blocknum * blocksize

    pervent = recv_size / totalsize
    percent_str = "%.2f%%" % (pervent * 100)
    n = round(pervent * 50)
    s = ('#' * n).ljust(50, '-')
    print('\r' + percent_str + '[' + s + ']' + speed_str, flush=True, end='')


def load_pickle(f):
    version = platform.python_version_tuple()
    if version[0] == '2':
        return pickle.load(f)
    elif version[0] == '3':
        return pickle.load(f, encoding='latin1')
    raise ValueError("invalid python version: {}".format(version))


def load_mnist(save_path=None):
    mnist_url = 'http://deeplearning.net/data/mnist/mnist.pkl.gz'
    if save_path is None:
        save_path = os.path.join(file_dir_path, 'data')
    if not os.path.exists(save_path):
        os.mkdir(save_path)
    save_path = os.path.join(save_path, 'mnist.pkl.gz')
    if os.path.exists(save_path):
        print(f'{save_path} alread exists!')
    else:
        print('>> Downloading MNIST Dataset')
        global start_time
        start_time = time.time()
        urllib.request.urlretrieve(mnist_url, save_path, schedule)
        print('Successfully downloaded')
    with gzip.open(save_path, 'rb') as f:
        train_set, val_set, test_set = load_pickle(f)
        return train_set, val_set, test_set
